@@ -9,13 +9,15 @@
 #if __GLASGOW_HASKELL__ >= 706
 {-# LANGUAGE DeriveGeneric #-}
 #endif
-#if __GLASGOW_HASKELL__ >= 710
+#if __GLASGOW_HASKELL__ >= 710 && __GLASGOW_HASKELL__ < 802
 {-# LANGUAGE AutoDeriveTypeable #-}
 #endif
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Data.FingerTree
--- Copyright   :  (c) Ross Paterson, Ralf Hinze 2006
+-- Copyright   :  Ross Paterson and Ralf Hinze 2006,
+--                Ross Paterson 2006-2022,
+--                James Cranch 2021
 -- License     :  BSD-style
 -- Maintainer  :  R.Paterson@city.ac.uk
 -- Stability   :  experimental
@@ -68,6 +70,8 @@ module Data.FingerTree (
     reverse,
     -- ** Maps
     fmap', fmapWithPos, fmapWithContext, unsafeFmap,
+    -- ** Folds
+    foldlWithPos, foldrWithPos, foldlWithContext, foldrWithContext,
     -- ** Traversals
     traverse', traverseWithPos, traverseWithContext, unsafeTraverse,
     -- * Example
@@ -287,39 +291,40 @@ fmapWithPos f = mapWPTree f mempty
 mapWPTree :: (Measured v1 a1, Measured v2 a2) =>
     (v1 -> a1 -> a2) -> v1 -> FingerTree v1 a1 -> FingerTree v2 a2
 mapWPTree _ _ Empty = Empty
-mapWPTree f v (Single x) = Single (f v x)
-mapWPTree f v (Deep _ pr m sf) =
-    deep (mapWPDigit f v pr)
-         (mapWPTree (mapWPNode f) vpr m)
-         (mapWPDigit f vm sf)
+mapWPTree f vl (Single x) = Single (f vl x)
+mapWPTree f vl (Deep _ pr m sf) =
+    deep (mapWPDigit f vl pr)
+         (mapWPTree (mapWPNode f) vlp m)
+         (mapWPDigit f vlpm sf)
   where
-    vpr     =  v    `mappend`  measure pr
-    vm      =  vpr  `mappend`  measure m
+    vlp     =  vl `mappend` measure pr
+    vlpm    =  vlp `mappend` measure m
 
 mapWPNode :: (Measured v1 a1, Measured v2 a2) =>
     (v1 -> a1 -> a2) -> v1 -> Node v1 a1 -> Node v2 a2
-mapWPNode f v (Node2 _ a b) = node2 (f v a) (f va b)
+mapWPNode f vl (Node2 _ a b) = node2 (f vl a) (f vla b)
   where
-    va      = v `mappend` measure a
-mapWPNode f v (Node3 _ a b c) = node3 (f v a) (f va b) (f vab c)
+    vla     =  vl `mappend` measure a
+mapWPNode f vl (Node3 _ a b c) = node3 (f vl a) (f vla b) (f vlab c)
   where
-    va      = v `mappend` measure a
-    vab     = va `mappend` measure b
+    va      =  vl `mappend` measure a
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
 
 mapWPDigit :: (Measured v a) => (v -> a -> b) -> v -> Digit a -> Digit b
-mapWPDigit f v (One a) = One (f v a)
-mapWPDigit f v (Two a b) = Two (f v a) (f va b)
+mapWPDigit f vl (One a) = One (f vl a)
+mapWPDigit f vl (Two a b) = Two (f vl a) (f vla b)
   where
-    va      = v `mappend` measure a
-mapWPDigit f v (Three a b c) = Three (f v a) (f va b) (f vab c)
+    vla     =  vl `mappend` measure a
+mapWPDigit f vl (Three a b c) = Three (f vl a) (f vla b) (f vlab c)
   where
-    va      = v `mappend` measure a
-    vab     = va `mappend` measure b
-mapWPDigit f v (Four a b c d) = Four (f v a) (f va b) (f vab c) (f vabc d)
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+mapWPDigit f vl (Four a b c d) = Four (f vl a) (f vla b) (f vlab c) (f vlabc d)
   where
-    va      = v `mappend` measure a
-    vab     = va `mappend` measure b
-    vabc    = vab `mappend` measure c
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+    vlabc   =  vlab `mappend` measure c
 
 -- | Map all elements of the tree with a function that also takes the
 -- measure of the prefix to the left and of the suffix to the right of
@@ -347,39 +352,41 @@ mapWCTree f vl (Deep _ pr m sf) vr =
 
 mapWCNode :: (Measured v1 a1, Measured v2 a2) =>
     (v1 -> a1 -> v1 -> a2) -> v1 -> Node v1 a1 -> v1 -> Node v2 a2
-mapWCNode f vl (Node2 _ a b) vr = node2 (f vl a vb) (f va b vr)
+mapWCNode f vl (Node2 _ a b) vr = node2 (f vl a vbr) (f vla b vr)
   where
-    va      = vl `mappend` measure a
-    vb      = measure b `mappend` vr
-mapWCNode f vl (Node3 _ a b c) vr = node3 (f vl a vbc) (f va b vc) (f vab c vr)
+    vla     =  vl `mappend` measure a
+    vbr     =  measure b `mappend` vr
+mapWCNode f vl (Node3 _ a b c) vr =
+    node3 (f vl a vbcr) (f vla b vcr) (f vlab c vr)
   where
-    va      = vl `mappend` measure a
-    vab     = va `mappend` measure b
-    vbc     = measure b `mappend` vc
-    vc      = measure c `mappend` vr
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+    vcr     =  measure c `mappend` vr
+    vbcr    =  measure b `mappend` vcr
 
 mapWCDigit ::
     (Measured v a) => (v -> a -> v -> b) -> v -> Digit a -> v -> Digit b
 mapWCDigit f vl (One a) vr = One (f vl a vr)
-mapWCDigit f vl (Two a b) vr = Two (f vl a vb) (f va b vr)
+mapWCDigit f vl (Two a b) vr = Two (f vl a vbr) (f vla b vr)
   where
-    va      = vl `mappend` measure a
-    vb      = measure b `mappend` vr
-mapWCDigit f vl (Three a b c) vr = Three (f vl a vbc) (f va b vc) (f vab c vr)
+    vla     =  vl `mappend` measure a
+    vbr     =  measure b `mappend` vr
+mapWCDigit f vl (Three a b c) vr =
+    Three (f vl a vbcr) (f vla b vcr) (f vlab c vr)
   where
-    va      = vl `mappend` measure a
-    vab     = va `mappend` measure b
-    vbc     = measure b `mappend` vc
-    vc      = measure c `mappend` vr
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+    vcr     =  measure c `mappend` vr
+    vbcr    =  measure b `mappend` vcr
 mapWCDigit f vl (Four a b c d) vr =
-    Four (f vl a vbcd) (f va b vcd) (f vab c vd) (f vabc d vr)
+    Four (f vl a vbcdr) (f vla b vcdr) (f vlab c vdr) (f vlabc d vr)
   where
-    va      = vl `mappend` measure a
-    vab     = va `mappend` measure b
-    vabc    = vab `mappend` measure c
-    vbcd    = measure b `mappend` vcd
-    vcd     = measure c `mappend` vd
-    vd      = measure d `mappend` vr
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+    vlabc   =  vlab `mappend` measure c
+    vdr     =  measure d `mappend` vr
+    vcdr    =  measure c `mappend` vdr
+    vbcdr   =  measure b `mappend` vcdr
 
 -- | Like 'fmap', but safe only if the function preserves the measure.
 unsafeFmap :: (a -> b) -> FingerTree v a -> FingerTree v b
@@ -392,6 +399,224 @@ unsafeFmapNode :: (a -> b) -> Node v a -> Node v b
 unsafeFmapNode f (Node2 v a b) = Node2 v (f a) (f b)
 unsafeFmapNode f (Node3 v a b c) = Node3 v (f a) (f b) (f c)
 
+-- | Fold the tree from the left with a function that also takes the
+-- measure of the prefix to the left of the element.
+--
+-- @since 0.1.5.0
+foldlWithPos :: (Measured v a) =>
+    (b -> v -> a -> b) -> b -> FingerTree v a -> b
+foldlWithPos f z = foldlWPTree f z mempty
+
+foldlWPTree :: (Measured v a) =>
+    (b -> v -> a -> b) -> b -> v -> FingerTree v a -> b
+foldlWPTree _ z _ Empty = z
+foldlWPTree f z vl (Single x) = f z vl x
+foldlWPTree f z vl (Deep _ pr m sf) = zpms
+  where
+    vlp     =  vl `mappend` measure pr
+    vlpm    =  vlp `mappend` measure m
+    zp      =  foldlWPDigit f z vl pr
+    zpm     =  foldlWPTree (foldlWPNode f) zp vlp m
+    zpms    =  foldlWPDigit f zpm vlpm sf
+
+foldlWPNode :: (Measured v a) =>
+    (b -> v -> a -> b) -> b -> v -> Node v a -> b
+foldlWPNode f z vl (Node2 _ a b) = f (f z vl a) vla b
+  where
+    vla     =  vl `mappend` measure a
+foldlWPNode f z vl (Node3 _ a b c) = f (f (f z vl a) vla b) vlab c
+  where
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+
+foldlWPDigit :: (Measured v a) =>
+    (b -> v -> a -> b) -> b -> v -> Digit a -> b
+foldlWPDigit f z vl (One a) = f z vl a
+foldlWPDigit f z vl (Two a b) = f (f z vl a) vla b
+  where
+    vla     =  vl `mappend` measure a
+foldlWPDigit f z vl (Three a b c) = f (f (f z vl a) vla b) vlab c
+  where
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+foldlWPDigit f z vl (Four a b c d) = f (f (f (f z vl a) vla b) vlab c) vlabc d
+  where
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+    vlabc   =  vlab `mappend` measure c
+
+-- | Fold the tree from the right with a function that also takes the
+-- measure of the prefix to the left of the element.
+--
+-- @since 0.1.5.0
+foldrWithPos :: (Measured v a) =>
+    (v -> a -> b -> b) -> b -> FingerTree v a -> b
+foldrWithPos f z = foldrWPTree f z mempty
+
+foldrWPTree :: (Measured v a) =>
+    (v -> a -> b -> b) -> b -> v -> FingerTree v a -> b
+foldrWPTree _ z _ Empty = z
+foldrWPTree f z vl (Single x) = f vl x z
+foldrWPTree f z vl (Deep _ pr m sf) = zpms
+  where
+    vlp     =  vl `mappend` measure pr
+    vlpm    =  vlp `mappend` measure m
+    zpms    =  foldrWPDigit f zms vl pr
+    zms     =  foldrWPTree (foldrWPNode f) zs vlp m
+    zs      =  foldrWPDigit f z vlpm sf
+
+-- different argument order for convenience
+foldrWPNode :: (Measured v a) =>
+    (v -> a -> b -> b) -> v -> Node v a -> b -> b
+foldrWPNode f vl (Node2 _ a b) z = f vl a (f vla b z)
+  where
+    vla     =  vl `mappend` measure a
+foldrWPNode f vl (Node3 _ a b c) z = f vl a (f vla b (f vlab c z))
+  where
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+
+foldrWPDigit :: (Measured v a) =>
+    (v -> a -> b -> b) -> b -> v -> Digit a -> b
+foldrWPDigit f z vl (One a) = f vl a z
+foldrWPDigit f z vl (Two a b) = f vl a (f vla b z)
+  where
+    vla     =  vl `mappend` measure a
+foldrWPDigit f z vl (Three a b c) = f vl a (f vla b (f vlab c z))
+  where
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+foldrWPDigit f z vl (Four a b c d) = f vl a (f vla b (f vlab c (f vlabc d z)))
+  where
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+    vlabc   =  vlab `mappend` measure c
+
+-- | Fold the tree from the left with a function that also takes the
+-- measure of the prefix to the left of the element and the measure of
+-- the suffix to the right of the element.
+--
+-- @since 0.1.5.0
+foldlWithContext :: (Measured v a) =>
+    (b -> v -> a -> v -> b) -> b -> FingerTree v a -> b
+foldlWithContext f z t = foldlWCTree f z mempty t mempty
+
+foldlWCTree :: (Measured v a) =>
+    (b -> v -> a -> v -> b) -> b -> v -> FingerTree v a -> v -> b
+foldlWCTree _ z _ Empty _ = z
+foldlWCTree f z vl (Single x) vr = f z vl x vr
+foldlWCTree f z vl (Deep _ pr m sf) vr = zpms
+  where
+    vlp     =  vl `mappend` measure pr
+    vlpm    =  vlp `mappend` vm
+    vmsr    =  vm `mappend` vsr
+    vsr     =  measure sf `mappend` vr
+    vm      =  measure m
+    zp      =  foldlWCDigit f z vl pr vmsr
+    zpm     =  foldlWCTree (foldlWCNode f) zp vlp m vsr
+    zpms    =  foldlWCDigit f zpm vlpm sf vr
+
+foldlWCNode :: (Measured v a) =>
+    (b -> v -> a -> v -> b) -> b -> v -> Node v a -> v -> b
+foldlWCNode f z vl (Node2 _ a b) vr = f (f z vl a vbr) vla b vr
+  where
+    vla     =  vl `mappend` measure a
+    vbr     =  measure b `mappend` vr
+foldlWCNode f z vl (Node3 _ a b c) vr =
+    f (f (f z vl a vbcr) vla b vcr) vlab c vr
+  where
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+    vcr     =  measure c `mappend` vr
+    vbcr    =  measure b `mappend` vcr
+
+foldlWCDigit :: (Measured v a) =>
+    (b -> v -> a -> v -> b) -> b -> v -> Digit a -> v -> b
+foldlWCDigit f z vl (One a) vr = f z vl a vr
+foldlWCDigit f z vl (Two a b) vr = f (f z vl a vbr) vla b vr
+  where
+    vla     =  vl `mappend` measure a
+    vbr     =  measure b `mappend` vr
+foldlWCDigit f z vl (Three a b c) vr =
+    f (f (f z vl a vbcr) vla b vcr) vlab c vr
+  where
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+    vcr     =  measure c `mappend` vr
+    vbcr    =  measure b `mappend` vcr
+foldlWCDigit f z vl (Four a b c d) vr =
+    f (f (f (f z vl a vbcdr) vla b vcdr) vlab c vdr) vlabc d vr
+  where
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+    vlabc   =  vlab `mappend` measure c
+    vdr     =  measure d `mappend` vr
+    vcdr    =  measure c `mappend` vdr
+    vbcdr   =  measure b `mappend` vcdr
+
+-- | Fold the tree from the right with a function that also takes the
+-- measure of the prefix to the left of the element and the measure of
+-- the suffix to the right of the element.
+--
+-- @since 0.1.5.0
+foldrWithContext :: (Measured v a) =>
+    (v -> a -> v -> b -> b) -> b -> FingerTree v a -> b
+foldrWithContext f z t = foldrWCTree f z mempty t mempty
+
+foldrWCTree :: (Measured v a) =>
+    (v -> a -> v -> b -> b) -> b -> v -> FingerTree v a -> v -> b
+foldrWCTree _ z _ Empty _ = z
+foldrWCTree f z vl (Single x) vr = f vl x vr z
+foldrWCTree f z vl (Deep _ pr m sf) vr = zpms
+  where
+    vlp     =  vl `mappend` measure pr
+    vlpm    =  vlp `mappend` vm
+    vmsr    =  vm `mappend` vsr
+    vsr     =  measure sf `mappend` vr
+    vm      =  measure m
+    zpms    =  foldrWCDigit f zms vl pr vmsr
+    zms     =  foldrWCTree (foldrWCNode f) zs vlp m vsr
+    zs      =  foldrWCDigit f z vlpm sf vr
+
+-- different argument order for convenience
+foldrWCNode :: (Measured v a) =>
+    (v -> a -> v -> b -> b) -> v -> Node v a -> v -> b -> b
+foldrWCNode f vl (Node2 _ a b) vr z = f vl a vbr (f vla b vr z)
+  where
+    vla     =  vl `mappend` measure a
+    vbr     =  measure b `mappend` vr
+foldrWCNode f vl (Node3 _ a b c) vr z =
+    f vl a vbcr (f vla b vcr (f vlab c vr z))
+  where
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+    vcr     =  measure c `mappend` vr
+    vbcr    =  measure b `mappend` vcr
+
+foldrWCDigit :: (Measured v a) =>
+    (v -> a -> v -> b -> b) -> b -> v -> Digit a -> v -> b
+foldrWCDigit f z vl (One a) vr = f vl a vr z
+foldrWCDigit f z vl (Two a b) vr = f vl a vbr (f vla b vr z)
+  where
+    vla     =  vl `mappend` measure a
+    vbr     =  measure b `mappend` vr
+foldrWCDigit f z vl (Three a b c) vr =
+    f vl a vbcr (f vla b vcr (f vlab c vr z))
+  where
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+    vcr     =  measure c `mappend` vr
+    vbcr    =  measure b `mappend` vcr
+foldrWCDigit f z vl (Four a b c d) vr =
+    f vl a vbcdr (f vla b vcdr (f vlab c vdr (f vlabc d vr z)))
+  where
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+    vlabc   =  vlab `mappend` measure c
+    vdr     =  measure d `mappend` vr
+    vcdr    =  measure c `mappend` vdr
+    vbcdr   =  measure b `mappend` vcdr
+
 -- | Like 'traverse', but with constraints on the element types.
 traverse' :: (Measured v1 a1, Measured v2 a2, Applicative f) =>
     (a1 -> f a2) -> FingerTree v1 a1 -> f (FingerTree v2 a2)
@@ -481,41 +706,41 @@ traverseWCTree f vl (Deep _ pr m sf) vr =
 
 traverseWCNode :: (Measured v1 a1, Measured v2 a2, Applicative f) =>
     (v1 -> a1 -> v1 -> f a2) -> v1 -> Node v1 a1 -> v1 -> f (Node v2 a2)
-traverseWCNode f vl (Node2 _ a b) vr = node2 <$> f vl a vb <*> f va b vr
+traverseWCNode f vl (Node2 _ a b) vr = node2 <$> f vl a vbr <*> f vla b vr
   where
-    va      = vl `mappend` measure a
-    vb      = measure a `mappend` vr
+    vla     =  vl `mappend` measure a
+    vbr     =  measure b `mappend` vr
 traverseWCNode f vl (Node3 _ a b c) vr =
-    node3 <$> f vl a vbc <*> f va b vc <*> f vab c vr
+    node3 <$> f vl a vbcr <*> f vla b vcr <*> f vlab c vr
   where
-    va      = vl `mappend` measure a
-    vab     = va `mappend` measure b
-    vc      = measure c `mappend` vr
-    vbc     = measure b `mappend` vc
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+    vcr     =  measure c `mappend` vr
+    vbcr    =  measure b `mappend` vcr
 
 traverseWCDigit :: (Measured v a, Applicative f) =>
     (v -> a -> v -> f b) -> v -> Digit a -> v -> f (Digit b)
 traverseWCDigit f vl (One a) vr = One <$> f vl a vr
-traverseWCDigit f vl (Two a b) vr = Two <$> f vl a vb <*> f va b vr
+traverseWCDigit f vl (Two a b) vr = Two <$> f vl a vbr <*> f vla b vr
   where
-    va      = vl `mappend` measure a
-    vb      = measure a `mappend` vr
+    vla     =  vl `mappend` measure a
+    vbr     =  measure b `mappend` vr
 traverseWCDigit f vl (Three a b c) vr =
-    Three <$> f vl a vbc <*> f va b vc <*> f vab c vr
+    Three <$> f vl a vbcr <*> f vla b vcr <*> f vlab c vr
   where
-    va      = vl `mappend` measure a
-    vab     = va `mappend` measure b
-    vc      = measure c `mappend` vr
-    vbc     = measure b `mappend` vc
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+    vcr     =  measure c `mappend` vr
+    vbcr    =  measure b `mappend` vcr
 traverseWCDigit f vl (Four a b c d) vr =
-    Four <$> f vl a vbcd <*> f va b vcd <*> f vab c vd <*> f vabc d vr
+    Four <$> f vl a vbcdr <*> f vla b vcdr <*> f vlab c vdr <*> f vlabc d vr
   where
-    va      = vl `mappend` measure a
-    vab     = va `mappend` measure b
-    vabc    = vab `mappend` measure c
-    vd      = measure d `mappend` vr
-    vcd     = measure c `mappend` vd
-    vbcd    = measure b `mappend` vcd
+    vla     =  vl `mappend` measure a
+    vlab    =  vla `mappend` measure b
+    vlabc   =  vlab `mappend` measure c
+    vdr     =  measure d `mappend` vr
+    vcdr    =  measure c `mappend` vdr
+    vbcdr   =  measure b `mappend` vcdr
 
 -- | Like 'traverse', but safe only if the function preserves the measure.
 unsafeTraverse :: (Applicative f) =>
@@ -885,7 +1110,7 @@ addDigits4 m1 (Four a b c d) e f g h (Four i j k l) m2 =
 --
 -- @since 0.1.2.0
 data SearchResult v a
-    = Position (FingerTree v a) a (FingerTree v a)
+    = Position !(FingerTree v a) a !(FingerTree v a)
         -- ^ A tree opened at a particular element: the prefix to the
         -- left, the element, and the suffix to the right.
     | OnLeft
@@ -961,13 +1186,13 @@ searchTree :: (Measured v a) =>
 searchTree _ _ Empty _ = illegal_argument "searchTree"
 searchTree _ _ (Single x) _ = Split Empty x Empty
 searchTree p vl (Deep _ pr m sf) vr
-  | p vlp vmsr  =  let  Split l x r     =  searchDigit p vl pr vmsr
-                   in   Split (maybe Empty digitToTree l) x (deepL r m sf)
-  | p vlpm vsr  =  let  Split ml xs mr  =  searchTree p vlp m vsr
-                        Split l x r     =  searchNode p (vlp `mappend` measure ml) xs (measure mr `mappend` vsr)
-                   in   Split (deepR pr  ml l) x (deepL r mr sf)
-  | otherwise   =  let  Split l x r     =  searchDigit p vlpm sf vr
-                   in   Split (deepR pr  m  l) x (maybe Empty digitToTree r)
+  | p vlp vmsr = case searchDigit p vl pr vmsr of
+    Split l x r -> Split (maybe Empty digitToTree l) x (deepL r m sf)
+  | p vlpm vsr = case searchTree p vlp m vsr of
+    Split ml xs mr -> case searchNode p (vlp `mappend` measure ml) xs (measure mr `mappend` vsr) of
+        Split l x r -> Split (deepR pr ml l) x (deepL r mr sf)
+  | otherwise = case searchDigit p vlpm sf vr of
+    Split l x r -> Split (deepR pr m l) x (maybe Empty digitToTree r)
   where
     vlp     =  vl `mappend` measure pr
     vlpm    =  vlp `mappend` vm
@@ -1054,20 +1279,20 @@ takeUntil p  =  fst . split p
 dropUntil :: (Measured v a) => (v -> Bool) -> FingerTree v a -> FingerTree v a
 dropUntil p  =  snd . split p
 
-data Split t a = Split t a t
+data Split t a = Split !t a !t
 
 splitTree :: (Measured v a) =>
     (v -> Bool) -> v -> FingerTree v a -> Split (FingerTree v a) a
 splitTree _ _ Empty = illegal_argument "splitTree"
 splitTree _ _ (Single x) = Split Empty x Empty
 splitTree p i (Deep _ pr m sf)
-  | p vpr       =  let  Split l x r     =  splitDigit p i pr
-                   in   Split (maybe Empty digitToTree l) x (deepL r m sf)
-  | p vm        =  let  Split ml xs mr  =  splitTree p vpr m
-                        Split l x r     =  splitNode p (vpr `mappend` measure ml) xs
-                   in   Split (deepR pr  ml l) x (deepL r mr sf)
-  | otherwise   =  let  Split l x r     =  splitDigit p vm sf
-                   in   Split (deepR pr  m  l) x (maybe Empty digitToTree r)
+  | p vpr = case splitDigit p i pr of
+    Split l x r -> Split (maybe Empty digitToTree l) x (deepL r m sf)
+  | p vm = case splitTree p vpr m of
+    Split ml xs mr -> case splitNode p (vpr `mappend` measure ml) xs of
+        Split l x r -> Split (deepR pr  ml l) x (deepL r mr sf)
+  | otherwise = case splitDigit p vm sf of
+    Split l x r -> Split (deepR pr  m  l) x (maybe Empty digitToTree r)
   where
     vpr     =  i    `mappend`  measure pr
     vm      =  vpr  `mappend`  measure m
@@ -1156,7 +1381,8 @@ Particular abstract data types may be implemented by defining
 element types with suitable 'Measured' instances.
 
 (from section 4.5 of the paper)
-Simple sequences can be implemented using a 'Sum' monoid as a measure:
+Simple sequences can be implemented using a 'Data.Monoid.Sum' monoid
+as a measure:
 
 > newtype Elem a = Elem { getElem :: a }
 >
