@@ -6,7 +6,7 @@
 #if __GLASGOW_HASKELL__ >= 706
 {-# LANGUAGE DeriveGeneric #-}
 #endif
-#if __GLASGOW_HASKELL__ >= 710
+#if __GLASGOW_HASKELL__ >= 710 && __GLASGOW_HASKELL__ < 802
 {-# LANGUAGE AutoDeriveTypeable #-}
 #endif
 -----------------------------------------------------------------------------
@@ -163,7 +163,7 @@ empty = PQueue FT.empty
 singleton :: Ord k => k -> v -> PQueue k v
 singleton k v = PQueue (FT.singleton (Entry k v))
 
--- | /O(log n)/. Add a (priority, value) pair to the front of a priority queue.
+-- | /O(1)/. Add a (priority, value) pair to the front of a priority queue.
 --
 -- * @'insert' k v q = 'union' ('singleton' k v) q@
 --
